# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List
from typing_extensions import TypeAlias

from .category_resource_response import CategoryResourceResponse

__all__ = ["CategoryDeleteResourceResponse"]

CategoryDeleteResourceResponse: TypeAlias = List[CategoryResourceResponse]
